/*
 * Copyright (C) 2011 Atmark Techno, Inc. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/gpio.h>
#include <mach/common.h>
#include <mach/hardware.h>
#include <mach/mmc.h>
#include "armadillo400_iomux.h"
#include "board.h"
#include "mx25_pins.h"

static struct pad_desc __maybe_unused armadillo400_sdhc1_pads[] = {
#if defined(CONFIG_ARMADILLO400_SDHC1_CON1)
	MX25_PAD_SD1_CMD__SD1_CMD(PAD_CTL_PUE | PAD_CTL_PUS_100K_UP | PAD_CTL_SRE_FAST),
	MX25_PAD_SD1_CLK__SD1_CLK(PAD_CTL_PUE | PAD_CTL_PUS_100K_UP | PAD_CTL_SRE_FAST),
	MX25_PAD_SD1_DATA0__SD1_DATA0(PAD_CTL_PUE | PAD_CTL_PUS_100K_UP | PAD_CTL_SRE_FAST),
	MX25_PAD_SD1_DATA1__SD1_DATA1(PAD_CTL_PUE | PAD_CTL_PUS_100K_UP | PAD_CTL_SRE_FAST),
	MX25_PAD_SD1_DATA2__SD1_DATA2(PAD_CTL_PUE | PAD_CTL_PUS_100K_UP | PAD_CTL_SRE_FAST),
	MX25_PAD_SD1_DATA3__SD1_DATA3(PAD_CTL_PUE | PAD_CTL_PUS_100K_UP | PAD_CTL_SRE_FAST),

	MX25_PAD_CTL_GRP_DSE_SDHC1(PAD_CTL_DSE_HIGH),

	/* SDHC1_CD */
	MX25_PAD_NFRB__GPIO_3_31(0),
	/* SDHC1_PWREN */
	MX25_PAD_NFRE_B__GPIO_3_27(),
#endif
};

static struct pad_desc __maybe_unused armadillo400_sdhc2_pads[] = {
#if defined(CONFIG_ARMADILLO400_CON9_16_SDHC2_CMD)
	MX25_PAD_CSI_D6__SD2_CMD(PAD_CTL_SRE_FAST),
#endif
#if defined(CONFIG_ARMADILLO400_CON9_18_SDHC2_CLK)
	MX25_PAD_CSI_D7__SD2_CLK(PAD_CTL_SRE_FAST),
#endif
#if defined(CONFIG_ARMADILLO400_CON9_21_SDHC2_DATA0)
	MX25_PAD_CSI_MCLK__SD2_DATA0(PAD_CTL_SRE_FAST),
#endif
#if defined(CONFIG_ARMADILLO400_CON9_22_SDHC2_DATA1)
	MX25_PAD_CSI_VSYNC__SD2_DATA1(PAD_CTL_SRE_FAST),
#endif
#if defined(CONFIG_ARMADILLO400_CON9_23_SDHC2_DATA2)
	MX25_PAD_CSI_HSYNC__SD2_DATA2(PAD_CTL_SRE_FAST),
#endif
#if defined(CONFIG_ARMADILLO400_CON9_24_SDHC2_DATA3)
	MX25_PAD_CSI_PIXCLK__SD2_DATA3(PAD_CTL_HYS | PAD_CTL_SRE_FAST),
#endif

	MX25_PAD_CTL_GRP_DSE_CSI(PAD_CTL_DSE_HIGH),

	/* SDHC2_CD */
#if defined(CONFIG_ARMADILLO400_CON9_17_SDHC2_CD)
	MX25_PAD_CSI_D9__GPIO_4_21(PAD_CTL_HYS | PAD_CTL_PKE | PAD_CTL_PUE | PAD_CTL_PUS_100K_UP),
#endif
	/* SDHC2_WP */
#if defined(CONFIG_ARMADILLO400_CON9_15_SDHC2_WP)
	MX25_PAD_CSI_D8__GPIO_1_7(PAD_CTL_HYS | PAD_CTL_PKE | PAD_CTL_PUE | PAD_CTL_PUS_100K_UP),
#endif
	/* SDHC2_PWREN */
#if defined(CONFIG_ARMADILLO400_CON9_1_SDHC2_PWREN)
	MX25_PAD_VSTBY_REQ__GPIO_3_17(PAD_CTL_PKE),
#endif
};

static void gpio_sdhc_request(int port)
{
	static int requested = 0;

	if (requested & (1 << port))
		return;

	switch (port) {
	case 0:
#if defined(CONFIG_ARMADILLO400_SDHC1_CON1)
		gpio_request(SDHC1_CLK_GPIO, "SDHC1_CLK");
		gpio_request(SDHC1_CD_GPIO, "SDHC1_CD");
#endif
		break;
	case 1:
#if defined(CONFIG_ARMADILLO400_SDHC2_CON9)
		gpio_request(SDHC2_CLK_GPIO, "SDHC2_CLK");
		gpio_request(SDHC2_CD_GPIO, "SDHC2_CD");
		gpio_request(SDHC2_WP_GPIO, "SDHC2_WP");
#endif
		break;
	default:
		break;
	}

	requested |= (1 << port);
}

static int __maybe_unused platform_sdhc_init(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);

	gpio_sdhc_request(pdev->id);

	switch (pdev->id) {
	case 0:
#if defined(CONFIG_ARMADILLO400_SDHC1_CON1)
		mxc_iomux_v3a_setup_multiple_pads(armadillo400_sdhc1_pads);

		/* Card-Detect pin */
		gpio_direction_input(SDHC1_CD_GPIO);
#endif
		break;
	case 1:
#if defined(CONFIG_ARMADILLO400_SDHC2_CON9)
		mxc_iomux_v3a_setup_multiple_pads(armadillo400_sdhc2_pads);

		/* Card-Detect pin */
		gpio_direction_input(SDHC2_CD_GPIO);

		/* Write-Protect pin */
		gpio_direction_input(SDHC2_WP_GPIO);
#endif
		break;
	default:
		/* This ID is not supported */
		break;
	}

	return 0;
}

static int __maybe_unused platform_sdhc_exit(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);

	gpio_sdhc_request(pdev->id);

	switch (pdev->id) {
	case 0:
#if defined(CONFIG_ARMADILLO400_SDHC1_CON1)
	{
		struct pad_desc sdhc1_inactive[] = {
			MX25_PAD_SD1_CLK__GPIO_2_24(PAD_CTL_PKE | PAD_CTL_PUE | PAD_CTL_PUS_100K_DOWN),
		};
		mxc_iomux_v3a_setup_multiple_pads(sdhc1_inactive);

		/* ENGcm02759: See "Chip Errata for the i.MX25" for details. */
		gpio_direction_output(SDHC1_CLK_GPIO, 0);

		gpio_direction_input(SDHC1_CD_GPIO);
	}
#endif
		break;
	case 1:
#if defined(CONFIG_ARMADILLO400_SDHC2_CON9)
	{
		struct pad_desc sdhc2_inactive[] = {
			MX25_PAD_CSI_D7__GPIO_1_6(PAD_CTL_PKE | PAD_CTL_PUE | PAD_CTL_PUS_100K_DOWN),
		};
		mxc_iomux_v3a_setup_multiple_pads(sdhc2_inactive);

		/* ENGcm02759: See "Chip Errata for the i.MX25" for details. */
		gpio_direction_output(SDHC2_CLK_GPIO, 0);

		gpio_direction_input(SDHC2_CD_GPIO);
		gpio_direction_input(SDHC2_WP_GPIO);
	}
#endif
		break;
	default:
		/* This ID is not supported */
		break;
	}

	return 0;
}

static unsigned int __maybe_unused sdhc_get_card_det_status(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);

	if (pdev->id == 0)
		return gpio_get_value(SDHC1_CD_GPIO);
	else
		return gpio_get_value(SDHC2_CD_GPIO);
}

#if defined(CONFIG_ARMADILLO400_SELECTED_SDHC1)
static struct mxc_mmc_platform_data armadillo400_sdhc1_data = {
	.ocr_mask = MMC_VDD_29_30 | MMC_VDD_32_33,
	.caps = MMC_CAP_4_BIT_DATA,
	.min_clk = 400000,
	.max_clk = 52000000,
	.card_inserted_state = 0,
	.status = sdhc_get_card_det_status,
	.clock_mmc = "esdhc_clk",
	.power_mmc = "sdhc1-power",

	.init = platform_sdhc_init,
	.exit = platform_sdhc_exit,
};

static struct resource mxcsdhc1_resources[] = {
	[0] = {
		.start = MX25_MMC_SDHC1_BASE_ADDR,
		.end = MX25_MMC_SDHC1_BASE_ADDR + SZ_4K - 1,
		.flags = IORESOURCE_MEM,
	},
	[1] = {
		.start = MX25_INT_SDHC1,
		.end = MX25_INT_SDHC1,
		.flags = IORESOURCE_IRQ,
	},
	[2] = {
		.start = IOMUX_TO_IRQ(MX25_PIN_NFRB),
		.end = IOMUX_TO_IRQ(MX25_PIN_NFRB),
		.flags = IORESOURCE_IRQ,
	},
};

static struct platform_device mxcsdhc1_device = {
	.name = "mxsdhci",
	.id = 0,
	.dev = {
		.platform_data = &armadillo400_sdhc1_data,
	},
	.num_resources = ARRAY_SIZE(mxcsdhc1_resources),
	.resource = mxcsdhc1_resources,
};
#endif

#if defined(CONFIG_ARMADILLO400_SELECTED_SDHC2)
static struct mxc_mmc_platform_data armadillo400_sdhc2_data = {
	.ocr_mask = MMC_VDD_29_30 | MMC_VDD_32_33,
	.caps = MMC_CAP_4_BIT_DATA,
	.min_clk = 400000,
	.max_clk = 52000000,
	.card_inserted_state = 0,
	.status = sdhc_get_card_det_status,
	.clock_mmc = "esdhc_clk",
#if defined(CONFIG_ARMADILLO400_SDHC2_PWREN_CON9_1)
	.power_mmc = "sdhc2-power",
#endif

	.init = platform_sdhc_init,
	.exit = platform_sdhc_exit,
};

static struct resource mxcsdhc2_resources[] = {
	[0] = {
		.start = MX25_MMC_SDHC2_BASE_ADDR,
		.end = MX25_MMC_SDHC2_BASE_ADDR + SZ_4K - 1,
		.flags = IORESOURCE_MEM,
	},
	[1] = {
		.start = MX25_INT_SDHC2,
		.end = MX25_INT_SDHC2,
		.flags = IORESOURCE_IRQ,
	},
	[2] = {
		.start = IOMUX_TO_IRQ(MX25_PIN_CSI_D9),
		.end = IOMUX_TO_IRQ(MX25_PIN_CSI_D9),
		.flags = IORESOURCE_IRQ,
	},
};

static struct platform_device mxcsdhc2_device = {
	.name = "mxsdhci",
	.id = 1,
	.dev = {
		.platform_data = &armadillo400_sdhc2_data,
	},
	.num_resources = ARRAY_SIZE(mxcsdhc2_resources),
	.resource = mxcsdhc2_resources,
};
#endif

void __init armadillo400_sdhc_init(void)
{
#if defined(CONFIG_ARMADILLO400_SELECTED_SDHC1)
	platform_device_register(&mxcsdhc1_device);
#endif
#if defined(CONFIG_ARMADILLO400_SELECTED_SDHC2)
	platform_device_register(&mxcsdhc2_device);
#endif
}
